

board_size = 8

SAMPLE_AZ_ARGS = {
    "num_net_channels": 512,
    "num_net_in_channels": 1,
    "net_dropout": 0.3,
    "net_action_size": board_size ** 2,
    "num_simulations": 1317,
    "self_play_games": 300,
    "num_iters": 50,
    "epochs": 500,
    "lr": 0.0032485504583772953,
    "max_buffer_size": 100_000,
    "num_pit_games": 40,
    "random_pit_freq": 3,
    "board_size": board_size,
    "batch_size": 256,
    "tau": 1,
    "arena_tau": 0.04139160592420218,
    "c": 1,
    "checkpoint_dir": None,
    "update_threshold": 0.6,
    "minimax_depth": 4,
    "show_tqdm": True,
    "num_workers": 5,
    "num_to_win": 5,
    "log_epsilon": 1e-9,
    "zero_tau_after": 5,
    "az_net_linear_input_size": 18432,
    "log_dir": "Logs",
    "pushbullet_token": None
}

TRAINED_AZ_NET_ARGS = {
    "num_net_channels": 512,
    "num_net_in_channels": 1,
    "net_dropout": 0.3,
    "net_action_size": board_size ** 2,
    "num_simulations": 1317,
    "self_play_games": 3,
    "num_iters": 50,
    "epochs": 320,
    "lr": 0.0032485504583772953,
    "max_buffer_size": 100_000,
    "num_pit_games": 40,
    "random_pit_freq": 3,
    "board_size": board_size,
    "batch_size": 128,
    "tau": 1.0,
    "arena_tau": 0,  # 0.04139160592420218
    "c": 1.15,
    "checkpoint_dir": None,
    "update_threshold": 0.6,
    "minimax_depth": 4,
    "show_tqdm": True,
    "num_workers": 5,
    "num_to_win": 4,
    "log_epsilon": 1.4165210108199043e-08,
    "log_dir": "Logs",
    "pushbullet_token": None
}

SAMPLE_MZ_ARGS = {
    "num_net_channels": 512,
    "num_net_out_channels": 256,
    "num_net_in_channels": 1,
    "net_dropout": 0.3,
    "net_action_size": 14,
    "net_latent_size": 36,
    "num_simulations": 240,
    "self_play_games": 5,
    "K": 5,
    "gamma": 0.997,
    "frame_buffer_size": 32,
    "frame_skip": 4,
    "num_steps": 400,
    "num_iters": 50,
    "epochs": 100,
    "lr": 0.001,
    "max_buffer_size": 70_000,
    "num_pit_games": 40,
    "random_pit_freq": 2,
    "board_size": board_size,
    "batch_size": 255,
    "tau": 1,
    "arena_tau": 1e-2,
    "c": 1,
    "c2": 19652,
    "alpha": 0.8,
    "checkpoint_dir": None,
    "update_threshold": 0.6,
    "minimax_depth": None,  # don't use with muzero
    "show_tqdm": False,
    "num_workers": 5,
    "num_to_win": 5,
    "log_epsilon": 1e-9,
    "zero_tau_after": 5,
    "beta": 1,
    "env_id": "ALE/Asteroids-v5",
    "pickle_dir": "Pickles/Data",
    "target_resolution": (96, 96),
    "az_net_linear_input_size": 8192,
    "log_dir": "Logs",
    "pushbullet_token": None
}
